//HavokCharacter.cpp
#include <Demos/demos.h>
#include <Physics/Dynamics/World/hkpWorld.h>
#include <Physics/Dynamics/Phantom/hkpSimpleShapePhantom.h>
#include <Physics/Utilities/CharacterControl/CharacterProxy/hkpCharacterProxy.h>
#include <Physics/Dynamics/Entity/hkpRigidBody.h>
#include <Physics/Utilities/CharacterControl/CharacterRigidBody/hkpCharacterRigidBody.h>
#include <Physics/Utilities/CharacterControl/CharacterRigidBody/hkpCharacterRigidBodyCinfo.h>
#include <Physics/Utilities/CharacterControl/CharacterRigidBody/hkpCharacterRigidBodyListener.h>
#include <Physics/Utilities/CharacterControl/StateMachine/hkpCharacterContext.h>
#include <Physics/Utilities/CharacterControl/StateMachine/hkpCharacterState.h>
#include <Physics/Utilities/CharacterControl/StateMachine/hkpCharacterStateManager.h>
#include <Physics/Utilities/CharacterControl/StateMachine/OnGround/hkpCharacterStateOnGround.h>
#include <Physics/Utilities/CharacterControl/StateMachine/InAir/hkpCharacterStateInAir.h>
#include <Physics/Utilities/CharacterControl/StateMachine/Jumping/hkpCharacterStateJumping.h>
#include <Physics/Utilities/CharacterControl/StateMachine/Climbing/hkpCharacterStateClimbing.h>


#include <Physics/Collide/Filter/Group/hkpGroupFilter.h>
#include <Physics/Collide/Shape/Convex/Capsule/hkpCapsuleShape.h>
#include <Physics/Collide/Shape/Convex/Cylinder/hkpCylinderShape.h>

#include "HavokCharacter.h"
#include "HavokManager.h"
#include <windows.h>
#include <assert.h>
#include <d3d9.h>
#include <d3dx9.h>


#define DEGTORAD(degree) ((D3DX_PI / 180.0f) * (degree)) 
#define RADTODEG(radian) ((180.0f / D3DX_PI) * (radian)) 

static const float fHAVOK_SCALE = 1.0f / 32.0f;
static const float fINVERSE_HAVOK_SCALE = 1.0f / fHAVOK_SCALE;

extern short GetKey( const short key );
extern void OutputDebugPrintf(char* fmt, ...);

HavokCharacter::HavokCharacter() 
: m_bActive( false )
, m_bCameraControl( true )
, m_bOldJump( false )
, m_pHavokRigidBody( NULL )
, m_characterContext( NULL )
, m_pStandShape( NULL )
, m_pCrouchShape( NULL )
{
	m_xMovement.x = 0;	m_xMovement.y = 0;

#ifdef USE_ZUP
	m_up.set( 0.0f, 0.0f, 1.0f );;
	m_forward.set( 0.0f, -1.0f, 0.0f );
#else
	m_up.set( 0.0f, 1.0f, 0.0f );;
	m_forward.set( 0.0f, 0.0f, 1.0f );
#endif
}
//--------------------------------------------------------------------------------
HavokCharacter::~HavokCharacter()
{

}
//--------------------------------------------------------------------------------
void HavokCharacter::Create( const hkVector4& position )
{
    if( m_bActive )
        return;

    m_bActive = true;

	hkpWorld* pWorld = HavokManager::Instance()->GetWorld();
	if( pWorld == NULL )
		return;

	hkVector4 scaledPosition = position;
	scaledPosition.mul4( fHAVOK_SCALE );

    // Define the shapes for the controller
#ifdef   USE_HAVOKDEMO
	hkVector4 va( 0.0f, 0.0f, CHARACTER_SCALE_RATIO * 0.2f );
	hkVector4 vb( 0.0f, 0.0f, CHARACTER_SCALE_RATIO * -0.2f );
	m_pStandShape = new hkpCapsuleShape( va, vb, CHARACTER_SCALE_RATIO * 0.497f );        
#else
    hkVector4 va( 0.0f, 0.0f, 0.1f );
    hkVector4 vb( 0.0f, 0.0f, -0.1f );
	m_pStandShape = new hkpCapsuleShape( va, vb, 1.1f );        
#endif
    va.setZero4();
    m_pCrouchShape = new hkpCapsuleShape( va, vb, 0.6f );


    // Construct a character rigid body
    hkpCharacterRigidBodyCinfo characterRigidBodyCInfo;
    characterRigidBodyCInfo.m_mass = 100.0f;
    characterRigidBodyCInfo.m_shape = m_pStandShape;

    characterRigidBodyCInfo.m_maxForce = 1000.0f;
    characterRigidBodyCInfo.m_up = m_up;
	 characterRigidBodyCInfo.m_position.set( position(0), position(1), position(2) );
    //characterRigidBodyCInfo.m_maxSlope = 65.0f * HK_REAL_DEG_TO_RAD;
	characterRigidBodyCInfo.m_maxSlope = 90.0f * HK_REAL_DEG_TO_RAD;
    characterRigidBodyCInfo.m_friction = 0;
	//characterRigidBodyCInfo.m_supportDistance = 0.1f * CHARACTER_SCALE_RATIO;
	//characterRigidBodyCInfo.m_unweldingHeightOffsetFactor = 5;
	characterRigidBodyCInfo.m_maxLinearVelocity = 1000;

    m_pHavokRigidBody = new hkpCharacterRigidBody( characterRigidBodyCInfo );

    hkpCharacterRigidBodyListener* pListener = new hkpCharacterRigidBodyListener();
    m_pHavokRigidBody->setListener( pListener );
    pListener->removeReference();

    pWorld->lock();
	pWorld->markForWrite();

    pWorld->addEntity( m_pHavokRigidBody->getRigidBody() );

    hkpCharacterState* pState = NULL;
    hkpCharacterStateManager* pManager = new hkpCharacterStateManager();

    pState = new hkpCharacterStateOnGround();
    pManager->registerState( pState, HK_CHARACTER_ON_GROUND );
    pState->removeReference();

    pState = new hkpCharacterStateInAir();
    pManager->registerState( pState, HK_CHARACTER_IN_AIR );
    pState->removeReference();

    pState = new hkpCharacterStateJumping();
    pManager->registerState( pState, HK_CHARACTER_JUMPING );
    pState->removeReference();

    pState = new hkpCharacterStateClimbing();
    pManager->registerState( pState, HK_CHARACTER_CLIMBING );
    pState->removeReference();

    m_characterContext = new hkpCharacterContext( pManager, HK_CHARACTER_ON_GROUND );
    pManager->removeReference();

    m_characterContext->setCharacterType( hkpCharacterContext::HK_CHARACTER_RIGIDBODY );
	m_characterContext->setFilterParameters(0.9f, 400, 400.0f);

	pWorld->unmarkForWrite();
    pWorld->unlock();
}
//--------------------------------------------------------------------------------
void HavokCharacter::Destroy()
{
    if( !m_bActive )
        return;

    m_bActive = false;

    hkpWorld* pWorld = HavokManager::Instance()->GetWorld();
    if( pWorld == NULL )
        return;

    m_pHavokRigidBody->removeReference();
    m_pStandShape->removeReference();
    m_pCrouchShape->removeReference();
    delete m_characterContext;
    m_pHavokRigidBody = NULL;
}
//--------------------------------------------------------------------------------
void HavokCharacter::Update( const float fTimestep )
{
#ifdef USE_HAVOKDIRECT_INPUT
	hkpWorld* pWorld = HavokManager::Instance()->GetWorld();
	if( pWorld == NULL )
		return;

    if( m_pHavokRigidBody == NULL )
        return;

    float fDX = 0.0f;
    float fDZ = 0.0f;
	float fStepWidth = CHARACTER_SCALE_RATIO * 0.5f;
	
	if( GetKey( VK_UP ) )
    {
        fDZ += fStepWidth;
    }

	if( GetKey( VK_DOWN ) )
    {
        fDZ -= fStepWidth;
    }

    if( GetKey( VK_LEFT ) )
    {
        fDX -= fStepWidth;
    }

    if( GetKey( VK_RIGHT ) )
    {
        fDX += fStepWidth;
    }

    pWorld->lock();
    pWorld->markForWrite();

	hkpCharacterOutput	output;
	hkpRigidBody* const pRigidBody = m_pHavokRigidBody->getRigidBody();

	// Calculate and set current acceleration for mass factor modifier
	{
		// calculate and set acceleration for mass factor modifier
		hkVector4 currentVel; 
		currentVel = pRigidBody->getLinearVelocity();

		hkVector4 currentAcc; 
		currentAcc.setSub4(output.m_velocity, currentVel);
		currentAcc.mul4( 1/fTimestep );

		m_pHavokRigidBody->setLinearAccelerationToMassModifier(currentAcc);
	}

	float down = 0;
	if( IsSupported() == false )
		down = -5;
	output.m_velocity = hkVector4( fDX, fDZ, down );
    m_pHavokRigidBody->setLinearVelocity( output.m_velocity, fTimestep );

 	pWorld->unmarkForWrite();
    pWorld->unlock();

	Vector3 pos;
	GetPosition( pos );

#else
	hkpWorld* pWorld = HavokManager::Instance()->GetWorld();
	if( pWorld == NULL )
		return;

    if( m_pHavokRigidBody == NULL )
        return;

    const bool bJump = m_bCameraControl && ( !m_bOldJump && GetKey( VK_SPACE ) );
    const bool bCrouch = m_bCameraControl && GetKey( 'C' );
    m_bOldJump = m_bCameraControl && GetKey( VK_SPACE );

    pWorld->lock();
    pWorld->markForWrite();


	hkVector4 p = m_pHavokRigidBody->getPosition();
	//OutputDebugPrintf( "Character Position : %f, %f, %f \n", p(0), p(1), p(2) );
	//ġ ̵
	if( GetKey( 'P' ) )
	{
		//SetPosition( -25.229128f, -54.980015f, 15.400537f );
		SetPosition( -2.340109f, -156.680710f, 47.019413f ); 
	}

    // TODO - the input code belongs in a player class or something
    hkpCharacterInput	xInput;
    hkpCharacterOutput	output;

    float fDX = 0.0f;
    float fDZ = 0.0f;
	float fStepWidth = CHARACTER_SCALE_RATIO * 0.5f;
	
	if( m_bCameraControl )
    {
        if( GetKey( VK_UP ) )
        {
            fDZ += fStepWidth;
        }

		if( GetKey( VK_DOWN ) )
        {
            fDZ -= fStepWidth;
        }

        if( GetKey( VK_LEFT ) )
        {
            fDX -= fStepWidth;
        }

        if( GetKey( VK_RIGHT ) )
        {
            fDX += fStepWidth;
        }
    }
    else
    {
		fDX = m_xMovement.x;
        fDZ = m_xMovement.y;
    }

    xInput.m_inputLR = fDX;
    xInput.m_inputUD = fDZ;

	xInput.m_wantJump = bJump;
    xInput.m_atLadder = false;

    xInput.m_up = m_up;

	xInput.m_forward = m_forward;
    xInput.m_forward.normalize3();
	
	hkQuaternion orient;
	orient.setAxisAngle( m_up, 0 );
	xInput.m_forward.setRotatedDir( orient, xInput.m_forward );

    hkStepInfo stepInfo;
    stepInfo.m_deltaTime = fTimestep;
    stepInfo.m_invDeltaTime = 1.0f / fTimestep;
	//stepInfo.m_endTime = HavokManager::Instance()->GetStopwatch()->getElapsedSeconds();
    xInput.m_stepInfo = stepInfo;

	hkVector4 characterGravity = m_up;
	float gravity = -16 * CHARACTER_SCALE_RATIO;
	characterGravity.mul4( hkVector4(  gravity, gravity, gravity ) );   
	xInput.m_characterGravity = characterGravity;
    
    hkpRigidBody* const pRigidBody = m_pHavokRigidBody->getRigidBody();

	xInput.m_velocity = pRigidBody->getLinearVelocity();
    xInput.m_position = pRigidBody->getPosition();

    m_pHavokRigidBody->checkSupport( stepInfo, xInput.m_surfaceInfo ); 

    const hkpCollidable* const pCollidable = pRigidBody->getCollidable();
	if( pRigidBody == NULL )
	{
		assert( 0 && "A character's Havok collidable pointer has been nulled. We are going to assume they aren't crouching." );
	}

    const bool bIsCrouched = pCollidable ? ( pRigidBody->getCollidable()->getShape() == m_pCrouchShape ) : false;

    if ( bIsCrouched && !bCrouch )
        pRigidBody->setShape( m_pStandShape );

    if ( !bIsCrouched && bCrouch )
        pRigidBody->setShape( m_pCrouchShape );

	// Apply the character state machine
    m_characterContext->update( xInput, output );

	// Calculate and set current acceleration for mass factor modifier
	{
		// calculate and set acceleration for mass factor modifier
		hkVector4 currentVel; 
		currentVel = pRigidBody->getLinearVelocity();

		hkVector4 currentAcc; 
		currentAcc.setSub4(output.m_velocity, currentVel);
		currentAcc.mul4( 1/fTimestep );

		m_pHavokRigidBody->setLinearAccelerationToMassModifier(currentAcc);
	}

    m_pHavokRigidBody->setLinearVelocity( output.m_velocity, fTimestep );

 	pWorld->unmarkForWrite();
    pWorld->unlock();

	GetPosition();
#endif
}

void HavokCharacter::LateUpdate()
{
    // TODO - third person cam here
    const bool bCrouch = ( GetKey( 'C' ) != 0 );
    const D3DXVECTOR3 cameraOffset = D3DXVECTOR3( 0.0f, bCrouch ? 0.4f : 0.8f, 0.0f );

    hkpWorld* pWorld = HavokManager::Instance()->GetWorld();

    if( pWorld == NULL )
        return;

    if( m_bCameraControl && m_pHavokRigidBody )
    {
        pWorld->lockReadOnly();
		pWorld->markForRead();

		//DDD
		/*
        GLToy_Camera::SetPosition(
            ( GLToy_Vector_3(
                m_pxHavokRigidBody->getPosition()( 0 ),
                m_pxHavokRigidBody->getPosition()( 1 ),
                m_pxHavokRigidBody->getPosition()( 2 ) ) + xCameraOffset ) * fINVERSE_HAVOK_SCALE );
		*/

		pWorld->unmarkForRead();
        pWorld->unlockReadOnly();
    }
}


void HavokCharacter::SetMovementDirection( const D3DXVECTOR2* xDirection )
{ 
	m_xMovement.x = xDirection->x; 
	m_xMovement.y = xDirection->y; 
}




hkTransform HavokCharacter::GetTransform()
{
	return m_pHavokRigidBody->getRigidBody()->getTransform();
}


void HavokCharacter::SetPosition( float x, float y, float z )
{
#if  0
    Destroy();
	hkVector4 vec( position->x, position->y, position->z, 0 );
    Create( vec );
#else
	hkpWorld* pWorld = HavokManager::Instance()->GetWorld();
    pWorld->lock();
    pWorld->markForWrite();

	hkVector4 pos( x, y, z );

	hkpRigidBody* pRigidBody = m_pHavokRigidBody->getRigidBody();
	pRigidBody->setPosition( pos );

 	pWorld->unmarkForWrite();
    pWorld->unlock();
#endif
}

bool HavokCharacter::GetPosition( Vector3& pos )
{
	//hkpWorld* pWorld = HavokManager::Instance()->GetWorld();
    //pWorld->lock();
    //pWorld->markForWrite();

	hkpRigidBody* pRigidBody = m_pHavokRigidBody->getRigidBody();
	if( pRigidBody == HK_NULL )
	{
 		//pWorld->unmarkForWrite();
		//pWorld->unlock();
		return false;
	}

	hkVector4 hkpos = pRigidBody->getPosition();
	pos.x = hkpos(0);
	pos.y = hkpos(1);
	pos.z = hkpos(2);

 	//pWorld->unmarkForWrite();
    //pWorld->unlock();
	return true;
}

bool HavokCharacter::IsSupported()
{
	hkReal timestep = 1.f / 60.f;
	// Check support
	hkpSurfaceInfo ground;

	hkStepInfo stepInfo; 
	stepInfo.m_deltaTime = timestep; 
	stepInfo.m_invDeltaTime = 1.f / timestep;

	hkpCharacterRigidBody* pCharRigidBody = m_pHavokRigidBody;
	//pCharRigidBody->getRigidBody()->markForRead();
	pCharRigidBody->checkSupport(stepInfo, ground);
	//pCharRigidBody->getRigidBody()->unmarkForRead();
	
	bool bRes = ( ground.m_supportedState != hkpSurfaceInfo::UNSUPPORTED );
	return bRes;
}